package com.yakindu.ea.examples.plugin.generators;

import static org.eclipse.uml2.uml.UMLPackage.Literals.CLASS;
import static org.eclipse.uml2.uml.UMLPackage.Literals.COMPONENT;
import static org.eclipse.uml2.uml.UMLPackage.Literals.DATA_TYPE;
import static org.eclipse.uml2.uml.UMLPackage.Literals.ENUMERATION;
import static org.eclipse.uml2.uml.UMLPackage.Literals.INTERFACE;
import static org.eclipse.uml2.uml.UMLPackage.Literals.MODEL;
import static org.eclipse.uml2.uml.UMLPackage.Literals.PACKAGE;
import static org.eclipse.uml2.uml.UMLPackage.Literals.PRIMITIVE_TYPE;
import static org.eclipse.uml2.uml.UMLPackage.Literals.PROFILE;
import static org.eclipse.uml2.uml.UMLPackage.Literals.STEREOTYPE;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.net.URL;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.PackageableElement;
import org.osgi.framework.Bundle;

import com.yakindu.bridges.ea.core.internal.utils.EAFileUtils;
import com.yakindu.ea.examples.plugin.EAExamplePlugin;

public class HtmlTreeGenerator {

	// some constants for files located in this plugin
	private static final Bundle BUNDLE = Platform.getBundle(EAExamplePlugin.PLUGIN_ID);
	private static final URL TEMPLATE = BUNDLE.getEntry("d3lib/html.template");
	private static final Map<String, URL> REQUIRED_FILES_MAP = new HashMap<String, URL>() {
		private static final long serialVersionUID = 1L;
		{
			put("d3.min.js", BUNDLE.getEntry("d3lib/d3.min.js"));
			put("D3_LICENSE", BUNDLE.getEntry("d3lib/D3_LICENSE"));
			put("background.png", BUNDLE.getEntry("d3lib/background.png"));
		}
	};

	// Element types for which specific icons are rendered in html template
	private static final Set<EClass> ELEMENT_TYPES_WITH_ICON = new HashSet<>(Arrays.asList(MODEL, PACKAGE, PROFILE,
			COMPONENT, CLASS, INTERFACE, DATA_TYPE, PRIMITIVE_TYPE, ENUMERATION, STEREOTYPE));

	private ResourceInfo resourceInfo;
	private HtmlTemplate htmlTemplate;

	public HtmlTreeGenerator(Resource resource) {
		if (resource == null)
			throw new IllegalArgumentException("resource must not be null");
		if (!resource.getURI().isPlatformResource())
			throw new IllegalArgumentException(
					"Currently, only platform:/resource URIs are supported, not: " + resource.getURI());
		resourceInfo = new ResourceInfo(resource);
		htmlTemplate = new HtmlTemplate(TEMPLATE);
	}

	public IFile generate() {
		final String data = generateData();
		final String html = htmlTemplate.prepareHtml(data, resourceInfo);
		final IFile htmlFile = getHtmlFile();
		save(htmlFile, html);
		ensureRequiredFilesExist(htmlFile.getParent());
		return htmlFile;
	}

	private String generateData() {
		final StringBuilder data = new StringBuilder();
		final Model[] models = resourceInfo.resource.getContents().stream() //
				.filter(Model.class::isInstance).map(Model.class::cast).toArray(Model[]::new);
		for (int i = 0; i < models.length; i++) {
			if (i > 0)
				data.append(",\n");
			generateData(models[i], data, 1);
		}
		return data.toString();
	}

	private void generateData(PackageableElement element, StringBuilder data, int depth) {
		final String elementName = resourceInfo.getElementName(element, depth);
		if (elementName == null)
			return; // skip unnamed elements
		final List<? extends PackageableElement> children = resourceInfo.getChildren(element);
		data.append("{'name':'");
		data.append(HtmlTemplate.escapeJsonString(elementName));
		if (ELEMENT_TYPES_WITH_ICON.contains(element.eClass())) {
			data.append("','type':'");
			data.append(element.eClass().getName());
		}
		data.append("'");
		if (!children.isEmpty()) {
			data.append(",\n'children':[");
			for (int i = 0; i < children.size(); i++) {
				final PackageableElement child = children.get(i);
				generateData(child, data, depth + 1);
				if (i < children.size() - 1) {
					data.append(",");
				}
			}
			data.append("\n]\n");
		}
		data.append("}");
	}

	private IFile getHtmlFile() {
		final URI htmlFileUri = resourceInfo.resource.getURI().trimFileExtension().appendFileExtension("html");
		return ResourcesPlugin.getWorkspace().getRoot().getFile(new Path(htmlFileUri.toPlatformString(true)));
	}

	private void save(IFile file, String contents) {
		try {
			final InputStream inputStream = new ByteArrayInputStream(contents.getBytes(file.getCharset()));
			if (file.exists()) {
				file.setContents(inputStream, true, true, null);
			} else {
				file.create(inputStream, true, null);
			}
		} catch (Exception e) {
			throw new RuntimeException("Failed to save file: " + file.getFullPath().toString(), e);
		}
	}

	private void ensureRequiredFilesExist(IContainer parent) {
		for (String fileName : REQUIRED_FILES_MAP.keySet()) {
			final IFile file = parent.getFile(new Path(fileName));
			if (!file.exists()) {
				EAFileUtils.copyURLFileToIFile(REQUIRED_FILES_MAP.get(fileName), file, null);
			}
		}
	}
}
